# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import argparse
import smax
import smax.log
import yaml


def generate_yaml(spec):
    # When dumping to yaml, hide the fields
    # beginning with underscore.
    def hide_underscores(dumper, o):
        r = {}
        for k, v in o.__dict__.items():
            if k.startswith("_"):
                continue
            r[k] = v
        return dumper.represent_mapping(
            "tag:yaml.org,2002:python/object:%s" % (o.__class__.__name__), r
        )

    yaml.add_multi_representer(object, hide_underscores)
    y = yaml.dump(spec, default_flow_style=False)
    return y


def main():
    parser = argparse.ArgumentParser(
        description="Translate a smax script to python.",
    )
    parser.add_argument(
        "--output",
        help="Write the output to the given filename",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Enable trace-level messages",
    )
    parser.add_argument(
        "--yaml",
        help="Yaml filename to write",
    )
    parser.add_argument(
        "input",
        help="input script; use '-' for standard input.",
    )
    args = parser.parse_args()

    smax.log.enable_trace = args.verbose

    filename = "/dev/stdin" if args.input == "-" else args.input
    source = smax.load_source(filename)
    spec, code = smax.translate(source, filename)
    output_filename = "/dev/stdout" if args.output is None else args.output
    with open(output_filename, "wt") as f:
        f.write(code)
    if args.yaml:
        y = generate_yaml(spec)
        with open(args.yaml, "wt") as f:
            f.write(y)


main()



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

# translate.py - Smax module which generates code.

import jinja2
import smax.log as log
import smax.parser
import time
import sys

environment = jinja2.Environment()


def machine(m):
    # we store some parameters in the state objects.
    for s in m.all_states():
        s._transition_methods = []
    t = environment.from_string(
        r"""
class {{ machine.name }}({{machine.superclass}}):
    # Make some printable strings to help diagnostics.
    {%- for state in machine.all_states() %}
    {{state.array_name}} = {{state.name_list|as_list}}
    {{state.full_name}} = "{{state.dot_name}}"
    {%- endfor %}{# state in machine.all_states() #}
    def __init__(self, reactor, debug_enable=False):
        self._reactor = reactor
        self._state = { }
        self._state_machine_debug_enable = debug_enable
        self._is_valid = False
        self._busy = False
    def _state_machine_debug(self, msg):
        if self._state_machine_debug_enable:
            print(
                "DEBUG -- %s, state=%s." % (msg, ",".join(self._state.keys()))
            )
    # for diagnostic only
    def _state_machine_enter(self, state_name):
        self._state_machine_debug("Entering %s" % ".".join(state_name))
    # for diagnostic only
    def _state_machine_exit(self, state_name):
        self._state_machine_debug("Exiting %s" % ".".join(state_name))
    # for diagnostic only
    def _state_machine_handle(self, state_name, event_name, *args):
        self._state_machine_debug(
            "%s handling %s" % (".".join(state_name), event_name)
        )
    def _state_machine_timeout(self, state_name, time_spec):
        self._state_machine_debug(
            "%s timed out after %s" % (".".join(state_name), time_spec)
        )
    def _state_machine_ignored(self, event_name, *args):
        self._state_machine_debug("Ignored %s" % event_name)
    def _state_machine_call_after_s(self, seconds, callback, context):
        self._state_machine_debug(
            "Scheduling %s timeout after %s seconds." % (context, seconds)
        )
        return self._reactor.after_s(seconds, callback)
    def _state_machine_call_after_ms(self, ms, callback, context):
        self._state_machine_debug(
            "Scheduling %s timeout after %s ms." % (context, ms)
        )
        return self._reactor.after_ms(ms, callback)
    def _state_machine_cancel_timeout(self, handle):
        return self._reactor.cancel_after(handle)
    def _record_state(self, state, timeouts):
        self._state[state] = timeouts
    def _unrecord_state(self, state):
        return self._state.pop(state)
    def _in_state(self, state):
        return state in self._state
    def start(self):
        if self._is_valid:
            raise RuntimeError("{{machine.name}} is already running")
        self._{{machine|munge("enter")}}()
        self._is_valid = True
    def end(self):
        self._{{machine|munge("unconfigure")}}()
    def call(self, cb, *args):
        self._reactor.call(cb, *args)
    # events
    {%- for ev in machine.event_list %}
    def {{ev.name}}({{ev.args|insert("self")|join(", ")}}):
        m = lambda self: self._{{machine.full_name}}_{{ev.name}}(
            {{ev.args|join(", ")}}
        )
        if self._busy:
            raise RuntimeError("{{machine.name}} recursive events (via a call to {{ev.name}}) is not supported.")
        try:
            self._busy = True
            return self._reactor._run_event(self, m)
        finally:
            self._busy = False
    {%- endfor %}{# ev in machine.event_list #}
    # states
    {%- for state in machine.all_states() %}
    def _{{state|munge("enter")}}(self):
        {%- if state.parent %}
        if not self._in_state(self.{{state.parent.full_name}}):
            self._{{state.parent|munge("configure")}}(
                [{{state|configure_list|join(", ")}}],
            )
        else:
            self._{{state|munge("configure")}}()
        {%- else %}{# state.parent #}
        self._{{state|munge("configure")}}()
        {%- endif %}{# state.parent #}
    def _{{state|munge("configure")}}(self, configurators=None):
        {%- set condition=["if"] %}
        {%- for or_peer in state.or_with %}
        {{condition[0]}} self._in_state(self.{{or_peer.full_name}}):
            self._{{or_peer|munge("unconfigure")}}()
        {%- set _ = condition.append("elif" if condition.pop() else "if") %}
        {%- endfor %}{# or_peer in self.or_with #}
        {%- if state.parent %}
        if not self._in_state(self.{{state.parent.full_name}}):
            self._state_machine_debug("Not in {{state.parent.full_name}}")
            self._{{state.parent|munge("configure")}}(
                [{{state|configure_list|join(", ")}}],
            )
            return
        {%- endif %}{# state.parent #}
        self._state_machine_enter(self.{{state.array_name}})
        {{-state|configure|indent(8)}}
        {%- if state.timeouts %}
        self._record_state(self.{{state.full_name}}, [
        {%- for t in state.timeouts %}
            self._state_machine_call_after_{{t.time_spec.scale}}(
                {{t.time_spec.timeout}},
                self._{{state|munge("timeout", loop.index0)}},
                self.{{state.full_name}},
            ),
        {%- endfor %}{# t in state.timeouts #}
            ])
        {%- else %}{# state.timeouts #}
        self._record_state(self.{{state.full_name}}, [])
        {%- endif %}{# state.timeouts #}
        if configurators is None:
            configurators = [{{state|child_list|join(", ")}}]
        for c in configurators:
            c()
        {{-state|transitions()|indent(8)}}
    def _{{state|munge("unconfigure")}}(self):
        {%- for sl in state.inner_states %}
        {%- set condition=["if"] %}
        {%- for s in sl %}
        {{condition[0]}} self._in_state(self.{{s.full_name}}):
            self._{{s|munge("unconfigure")}}()
        {%- set _ = condition.append("elif" if condition.pop() else "if") %}
        {%- endfor %}{# s in sl #}
        {%- endfor %}{# sl in state.inner_states #}
        self._state_machine_exit(self.{{state.array_name}})
        timeout_list = self._unrecord_state(self.{{state.full_name}})
        for t in timeout_list:
            self._state_machine_cancel_timeout(t)
        {{ state.exit|code("exit")|indent(8) }}
    {%- for event in machine.event_list %}
    {%- if event in state._events %}
    def _{{state.full_name}}_{{event.name}}(
        {{event.args|insert("self")|join(", ")}},
    ):
        if not self._is_valid:
            raise RuntimeError(
                "{{event.name}}: {{machine.name}} is not running",
            )
        r = False
        # Check inner states (if any)
        {%- for sl in state.inner_states %}
        {%- set condition=["if"] %}
        {%- for s in sl %}
        {%- if event in s._events %}
        {{condition[0]}} self._in_state(self.{{s.full_name}}):
            r = \
              self._{{s.full_name}}_{{event.name}}({{event.args|join(", ")}}) \
              or r
        {%- set _ = condition.append("elif" if condition.pop() else "if") %}
        {%- endif %}{# event in s._events #}
        {%- endfor %}{# s in sl #}
        {%- endfor %}{# sl in state.inner_states #}
        if r:
            return True
        {{-state|transitions(event)|indent(8)}}
        {%- for name, args in event.superclasses %}
        # superclass: {{name}} {{args}}
        if not r:
            r = self._{{machine.name}}_{{name}}({{args|join(", ")}})
        {%- endfor %}{# superclass in event.superclasses #}
        {%- if not state.parent %}
        if not r:
            self._state_machine_ignored(
                {{args|insert("\"%s\"" % event.name)|join(", ")}},
            )
        {%- endif %}{# not state.parent #}
        return r
    {%- endif %}{# event in state._events #}
    {%- endfor %}{# event in machine.event_list #}
    {%- for timeout in state.timeouts %}
    def _{{state|munge("timeout", loop.index0)}}(self):
        {%- if timeout.condition %}
        if not ({{timeout.condition}}):
            return
        {%- endif %}{# timeout.condition #}
        self._state_machine_timeout(
            self.{{state.array_name}},
            "{{timeout.time_spec.timeout}}{{timeout.time_spec.scale}}",
        )
        {{-timeout|goto|indent(8)}}
    {%- endfor %}{# timeout in state.timeouts #}
    {%- for transition in state._transition_methods %}
    {{- transition|transition_method|indent(4) }}
    {%- endfor %}{# transition in state._transition_methods #}
    {%- endfor %}{# state in machine.all_states() #}
"""
    )
    r = t.render(machine=m)
    return r


def transitions(state, event=None):
    t = environment.from_string(
        r"""
{%- if transition.condition %}
if {{transition.condition}}:
    self._state_machine_handle({{event_args|join(", ")}})
    {{-transition|goto|indent(4)}}
    return True
{%- else %}{# transition.condition #}
self._state_machine_handle({{event_args|join(", ")}})
{{-transition|goto|indent(0)}}
r = True
{%- endif %}{# transition.condition #}
"""
    )
    r = []
    already_has_default_transition = False
    for transition in state.transitions:
        if transition.event == event:
            event_args = ["self." + state.full_name + "_name"]
            if event:
                event_args.append('"%s"' % event.name)
                event_args.extend(['"%s"' % s for s in event.args])
            else:
                event_args.append(None)
                # there should only be at most one of these.
                assert not already_has_default_transition
                if not transition.condition:
                    already_has_default_transition = True
            r.append(
                t.render(
                    transition=transition,
                    state=state,
                    event_args=event_args,
                )
            )
    return "".join(r)


def configure(state):
    t = environment.from_string(
        r"""
{{state.enter|code("enter")}}
"""
    )
    return t.render(state=state)


def timeouts(state):
    r = []
    for n, t in enumerate(state.timeouts):
        r.append(
            "self._state_machine_call_after_%s(%s, self._%s)"
            % (
                t.time_spec.scale,
                t.time_spec.timeout,
                munge(state, "timeout", n),
            )
        )
    return r


def code(c, context):
    if not c or (len(c) == 0):
        return "pass # (no %s code was specified)" % (context,)
    return "\n".join(c)


def goto(transition):
    state = transition.state
    if transition not in state._transition_methods:
        state._transition_methods.append(transition)
    event = hasattr(transition, "event")
    args = []
    if event and transition.event:
        args.extend(transition.event.args)
    t = environment.from_string(
        r"""
{{transition_method}}({{args|join(", ")}})
"""
    )
    return t.render(
        transition=transition,
        args=args,
        transition_method="self._%s" % transition_name(transition),
    )


def transition_method(transition):
    event = hasattr(transition, "event")
    args = []
    if event:
        if transition.event:
            args.extend(transition.event.args)
    t = environment.from_string(
        r"""
def _{{transition|transition_name}}({{args|insert("self")|join(", ")}}):
    {%- if transition.unconfigure %}
    self._{{transition.state|munge("unconfigure")}}()
    {%- endif %}{# transition.unconfigure #}
    # transition code
    {{ transition.code|code("transition")|indent(4) }}
    {%- if transition.target_state %}
    self._{{transition.target_state|munge("configure")}}()
    {%- endif %}{# transition.target_state #}
"""
    )
    return t.render(transition=transition, args=args)


def munge(state, context, index=None):
    log.trace(
        "munge, state=%s, context=%s, index=%s."
        % (
            state.name,
            context,
            index,
        ),
    )
    r = [state.full_name, context]
    if index is not None:
        r.append("%s" % index)
    return "_".join(r)


def transition_name(transition):
    timeout = not hasattr(transition, "event")
    state = transition.state
    return "%s_%s_%s" % (
        state.full_name,
        "timedout" if timeout else "transition",
        transition.n,
    )


def insert(sequence, s):
    r = [s]
    r.extend(sequence)
    return r


# Returns a list of methods that the parent
# will call to properly set itself up.
def configure_list(state):
    sequence = []
    for sl in state.parent.inner_states:
        if state in sl:
            sequence.append("self._%s" % munge(state, "configure"))
            continue
        for s in sl:
            if s.start:
                sequence.append("self._%s" % munge(s, "configure"))
    return sequence


def child_list(state):
    sequence = []
    for sl in state.inner_states:
        for s in sl:
            if s.start:
                sequence.append("self._%s" % munge(s, "configure"))
    return sequence


def as_list(sequence):
    r = ['"%s"' % i for i in sequence]
    return "[%s]" % ",".join(r)


environment.filters["machine"] = machine
environment.filters["munge"] = munge
environment.filters["insert"] = insert
environment.filters["configure"] = configure
environment.filters["goto"] = goto
environment.filters["timeouts"] = timeouts
environment.filters["code"] = code
environment.filters["transitions"] = transitions
environment.filters["configure_list"] = configure_list
environment.filters["child_list"] = child_list
environment.filters["transition_name"] = transition_name
environment.filters["transition_method"] = transition_method
environment.filters["as_list"] = as_list


def parse(source, filename):
    scanner = smax.parser.Scanner(source, filename=filename)
    p = smax.parser.Parser(scanner)
    s = p.parse()
    spec = {
        "program_name": sys.argv[0],
        "run_timestamp": time.strftime("%Y-%m-%d %H:%M:%S"),
        "source_name": filename,
        "spec": s,
    }
    return spec


def generate_python(spec):
    # Generate the output.
    t = environment.from_string(
        r"""
# Generated by {{ program_name }} from {{ source_name }}.
{%- for s in spec %}
{#- Constant? #}
{%- if "constant" in s %}
{{ s.constant.name }} = {{ s.constant.value }}
{%- endif %}
{#- Import? #}
{%- if "import" in s %}
{{ s.import }}
{%- endif %}
{#- Machine? #}
{%- if "machine" in s %}
{{ s.machine|machine }}
{%- endif %}
{%- endfor %}
"""
    )
    s = t.render(spec) + "\n"
    return s


def translate(source, filename, yaml_filename=None):
    spec = parse(source, filename)
    code = generate_python(spec)
    return spec, code



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import types

from .parser import load_source
from .reactor import Reactor  # noqa: F401
from .asyncio_reactor import AsyncioReactor  # noqa: F401
from .select_reactor import SelectReactor  # noqa: F401
from .translate import parse, generate_python, translate  # noqa: F401


def compile_python(python_code, module_name="state_machine"):
    """
    Returns a module with the given name.
    """
    # Create the module we're return with
    m = types.ModuleType(module_name)
    exec(python_code, m.__dict__)
    # Normally you'd add the module to sys.modules
    # but we only want this module to be visible
    # to the caller.
    return m


def _load(filename):
    """We cache the compiled results from filename
    so you can get other machine class implementations
    from this same file quickly.
    """
    global smax_modules
    source, spec, python_code, module = smax_modules.get(
        filename, (None, None, None, None)
    )
    if source is None:
        source = load_source(filename)
        spec = parse(source, filename)
        python_code = generate_python(spec)
        module = compile_python(python_code)
    return source, spec, python_code, module


def load(filename, class_name, save_generated_python=lambda s: None):
    source, spec, python_code, module = _load(filename)
    save_generated_python(python_code)
    return module.__dict__[class_name]


def spec(filename):
    source, spec, python_code, module = _load(filename)
    return spec


smax_modules = {}



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import asyncio
import smax
import smax.log as log


class AsyncioReactor(smax.Reactor):
    update = b"U"

    def __init__(self, event_loop):
        self._signal_queue = asyncio.Queue()
        self._event_loop = event_loop
        super(AsyncioReactor, self).__init__()

    async def run(self):
        while True:
            timeout = self.sync()
            if self.done():
                return
            # timeout may be None
            log.trace("timeout=%s." % timeout)
            try:
                # we ignore the returned value here;
                # it's always self.update.
                await asyncio.wait_for(self._signal_queue.get(), timeout)
            except asyncio.exceptions.TimeoutError:
                pass

    def _signal(self):
        self._signal_queue.put_nowait(self.update)

    def _run_event(self, machine, ev):
        future = self._event_loop.create_future()
        self.call(self._do_run_event, future, machine, ev)
        return future

    def _do_run_event(self, future, machine, ev):
        try:
            r = ev(machine)
            future.set_result(r)
        except Exception as e:
            future.set_exception(e)



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import io
import pkgutil
import re
import smax.log as log
import yapps.grammar
import yapps.parsetree
import yapps.runtime


def load_source(
    filename,
    delimiter="%%",
    start_delimiter=None,
    end_delimiter=None,
):
    """
    Loads a file, replacing all the sections outside the '%%' delimited area
    with blank lines.  For example,
    ...
    ...       (replaced with blank lines)
    ...
    %%        (replaced with a blank line)
    ...
    ...       (left in the output)
    ...
    %%        (replaced with a blank line)
    ...
    ...       (replaced with blank lines)
    ...
    %%        (replaced with a blank line)
    ...
    ...       (more input that's copied to the output)
    ...
    %%        (replaced with a blank line)
    ...
    ...       (replaced with blank lines)
    ...
    This way, syntax errors in the delimited section show up
    with the right line numbers from the original source file.

    You can change the delimiters: Pass a parameter 'delimiter="XY"' to use XY
    instead of %% as delimiters or 'start_delimiter="%{", end_delimiter="%}"'
    to use different delimiters to start and end.
    """
    if start_delimiter is None:
        start_delimiter = delimiter
    if end_delimiter is None:
        end_delimiter = delimiter
    log.trace("load_source filename=%s" % (filename,))
    with open(filename, "rt") as f:
        return load_file(f, start_delimiter, end_delimiter)


def load_file(f, start_delimiter="%%", end_delimiter="%%"):
    lines = []
    while True:
        while True:
            # replace with blanks
            line = f.readline()
            if len(line) == 0:
                break
            lines.append("")
            line = line.rstrip()
            if line == start_delimiter:
                break
        if len(line) == 0:
            break
        while True:
            # copy to the output.
            line = f.readline()
            if len(line) == 0:
                break
            line = line.rstrip()
            if line == end_delimiter:
                break
            lines.append(line)
        if len(line) == 0:
            break
        lines.append("")
    spec = "\n".join(lines)
    return spec


r"""  # noqa: E501
%%

parser state_machine:

    # The scanner returns INDENT and DEDENT by measuring whitespace at the beginning
    # of the line.  To make yapps happy, we provide dummy text it can match,
    # then override the scanner to actually return these tokens as appropriate.
    # INDENTED_CODE fails if a dedent is found; otherwise returns
    # the complete line of text found with the current indent prefix removed.
    token INDENT: "faking out the scanner with something that shouldnt ever match, part 1"
    token DEDENT: "faking out the scanner with something that shouldnt ever match, part 2"
    token INDENTED_CODE: "faking out the scanner with something that shouldnt ever match, part 3"
    token MACHINE: "machine"
    token MACRO: "macro"
    token STATE: "state"
    token PASS: "pass"
    token ENTER: "enter"
    token EXIT: "exit"
    token IMPORT: "import"
    token IS: "is"
    token START: "\\*"
    token AND: "---"
    token TRANSITION: "->"
    token OPEN_BRACKET: r"\["
    token CLOSE_BRACKET: r"\]"
    token NONBRACKET: r"[^\[\]]*"
    token OPEN_PAREN: r"\("
    token CLOSE_PAREN: r"\)"
    token NONPAREN: r"[^\(\)]*"
    token MS: "ms"
    token S: "s"
    token EOF: "$"
    token UP: r"\^"
    token FLOAT0: r"[0-9]+\.[0-9]*"
    token FLOAT1: r"[0-9]*\.[0-9]+"
    token INT: r"[0-9]+"
    token NAME: r"[\w]+"
    token TOEOL: r".*"
    ignore: r"#.*"      # comments
    ignore: r"[ \r\t\n]+"

    rule machine_spec<<spec>>:
        (   machine<<spec>>
        |   macro<<spec>>
        |   constant<<spec>>
        |   import_<<spec>>      # 'import' is a keyword so we use this alias.
        )*
        EOF
        {{ return spec.spec() }}

    rule constant<<spec>>:
        NAME '=' TOEOL {{ return spec.constant(NAME.strip(), TOEOL.strip()) }}

    rule import_<<spec>>:
        IMPORT TOEOL {{ return spec.import_("import %s" % TOEOL.strip()) }}

    rule machine<<spec>>:
        {{ superclass = "object" }}
        MACHINE machine_name [ OPEN_PAREN NAME CLOSE_PAREN {{ superclass = NAME }} ] ':'
        {{ machine = spec.machine(machine_name, superclass) }}
        INDENT
            ( state_machine_list<<machine.context()>>
            | enter_clause              {{  machine.set_enter(enter_clause) }}
            | exit_clause               {{  machine.set_exit(exit_clause) }}
            | transition<<machine.context()>>
            )*
        DEDENT
        {{ machine.check() }}

    rule macro<<spec>>:
        MACRO NAME ':'
        {{ macro = spec.macro(NAME) }}
        INDENT
            ( transition<<macro>>
            | timeout<<macro>>
            | inner_states<<macro>>
            | enter_clause              {{  macro.set_enter(enter_clause) }}
            | exit_clause               {{  macro.set_exit(exit_clause) }}
            | default_transition<<macro>>
            | PASS
            )*
        DEDENT
        {{ macro.check() }}

    rule state_machine_list<<context>>:
        [ AND ] # this allows a '---' before the first state.
        state_list<<context>> {{ context.state_machine(state_list) }}
        ( AND state_list<<context>> {{ context.state_machine(state_list) }} )*

    rule state_list<<context>>:
        {{ states = [ ] }}
        ( state_decl<<context>> {{ states.append(state_decl) }} )+
        {{ return states }}

    rule state_decl<<context>>:
        start STATE
            state_name ':'
                {{ state = context.state(state_name, start) }}
        INDENT
            ( transition<<state>>
            | timeout<<state>>
            | inner_states<<state>>
            | enter_clause              {{  state.set_enter(enter_clause) }}
            | exit_clause               {{  state.set_exit(exit_clause) }}
            | default_transition<<state>>
            | PASS
            )*
        DEDENT
        {{ return state }}

    rule inner_states<<state>>:
        {{ inner_context = state.new_context() }}
        state_machine_list<<inner_context>>

    rule start:
        {{ r=False }}
        ( START {{ r=True }} )?
        {{ return r }}

    rule transition<<state>>:
        {{ event_args=[ ] }}
        {{ condition=None }}
        {{ state_target=None }}
        {{ code_clause=None }}
        {{ state_target=None }}
        {{ superclasses=[ ] }}
        event_name
            ( OPEN_PAREN event_args CLOSE_PAREN )?
            ( superclass<<superclasses>> )*
            ( OPEN_BRACKET condition CLOSE_BRACKET )?
            ( TRANSITION state_target<<state>> [ ':' code_clause ]
            | ':' (code_clause|PASS)
            )
        {{ state.add_transition(event_name, event_args, superclasses, condition, state_target, code_clause) }}

    rule superclass<<sl>>:
        {{ event_args = None }}
        IS NAME ( OPEN_PAREN event_args CLOSE_PAREN )?
            {{ sl.append( [NAME, event_args] ) }}

    rule event_args:
        {{ args = [] }}
        NAME {{ args.append(NAME) }}
        ( ',' NAME {{ args.append(NAME) }} )*
        {{ return args }}

    rule default_transition<<state>>:
        {{ condition=None }}
        {{ code_clause=None }}
        ( OPEN_BRACKET condition CLOSE_BRACKET )?
        TRANSITION state_target<<state>> [ ':' code_clause ]
        {{ state.default_transition(condition, state_target, code_clause) }}

    rule timeout<<state>>:
        {{ code_clause = None }}
        {{ condition = None }}
        {{ state_target = None }}
        time_spec<<state>>
        ( OPEN_BRACKET condition CLOSE_BRACKET )?
        [ TRANSITION state_target<<state>>] [':' code_clause]
        {{ state.add_timeout(time_spec, condition, state_target, code_clause) }}

    rule time_spec<<state>>:
        (   MS OPEN_PAREN expr CLOSE_PAREN {{ return state.timeout_ms(expr) }}
        |   S OPEN_PAREN expr CLOSE_PAREN {{ return state.timeout_s(expr) }}
        )

    rule float:
        (   INT {{ r = float(INT) }}
        |   FLOAT0 {{ r = float(FLOAT0) }}
        |   FLOAT1 {{ r = float(FLOAT1) }}
        )
        {{ return r }}

    rule condition:
        _condition<<[]>>
        {{ return "".join(_condition) }}

    rule _condition<<l>>:
        ( NONBRACKET {{ l.append(NONBRACKET) }} )?
        ( OPEN_BRACKET condition<<[]>> CLOSE_BRACKET
            {{ l.append("["); l.extend(condition); l.append("]") }}
        )*
        {{ return l }}

    rule expr:
        {{ l = [] }}
        ( _expr<<[]>> {{ l.append("".join(_expr)) }} )*
        {{ return "".join(l) }}

    rule _expr<<l>>:
        (   NONPAREN    {{ l.append(NONPAREN) }}
        |   OPEN_PAREN _expr<<[]>> CLOSE_PAREN
                {{ l.append("("); l.extend(_expr); l.append(")") }}
        )
        {{ return l }}

    rule enter_clause:
        ENTER ':' code_clause
        {{ return code_clause }}

    rule exit_clause:
        EXIT ':' code_clause
        {{ return code_clause }}

    rule code_clause:
        ( simple_code_clause {{ return simple_code_clause }}
        | indented_code_clause {{ return indented_code_clause }}
        )

    rule simple_code_clause:
        TOEOL {{ return [ TOEOL.strip() ] }}

    rule indented_code_clause:
        INDENT {{ r = [ ] }}
            ( INDENTED_CODE {{ r.append(INDENTED_CODE) }} )+
        DEDENT {{ return r }}

    rule event_name:
        NAME {{ return NAME }}

    rule machine_name:
        NAME {{ return NAME }}

    rule state_name:
        NAME {{ return NAME }}

    rule state_target<<state>>:
        {{ r = [ ] }}
        ( UP {{ r.append(UP) }} )*
        NAME  {{ r.append(NAME) }}
        ( r'\.' NAME {{ r.append(NAME) }} )*
        {{ return r }}

%%
"""


class Buffer:
    def __init__(self):
        self._buffer = []

    def write(self, msg):
        self._buffer.append(msg)


def load_parser():
    source_data = pkgutil.get_data(__name__, "parser.py").decode("utf-8")
    source_file = io.StringIO(source_data)
    source = load_file(source_file)
    scanner = yapps.grammar.ParserDescriptionScanner(source, filename=__file__)
    parser = yapps.grammar.ParserDescription(scanner)
    # monkey-patch the writer so we catch the python code
    t = yapps.runtime.wrap_error_reporter(parser, "Parser")
    t.output = Buffer()
    t.postparser = "\n\n"
    t.generate_output()
    parser_python = "".join(t.output._buffer)
    if False:
        with open("parser-out", "wt") as f:
            f.write(parser_python)
    # run the generated python code.
    exec(parser_python, globals())


load_parser()


class Scanner(state_machineScanner):  # noqa: F821
    def __init__(self, *args, **kwargs):
        super(Scanner, self).__init__(*args, **kwargs)
        self._indent = [0]
        self._eof = None
        self._spaces = re.compile("\n([ ]*)")
        self._blank_line = re.compile("\n([ ]*(#[^\n]*)?)\n")
        self._indented_code = re.compile("\n([ ]*)([^\n]*)")

    def token(self, restrict, context=None):
        log.trace("restrict=%s, pos=%s." % (restrict, self.get_pos()))
        # If they're looking for INDENT or DEDENT,
        # see if we're at the end-of-line followed by some
        # spaces; if the indent level changes, then pass that
        # token back up.  We'll consider EOF to be a DEDENT
        # if we're looking for dedent too.
        if ("DEDENT" in restrict) and (self.pos >= len(self.input)):
            token = yapps.runtime.Token("DEDENT", 0, self.get_pos())
            return token
        if ("INDENT" in restrict) or ("DEDENT" in restrict):
            while True:
                m = self._blank_line.match(self.input, self.pos)
                if m:
                    self.pos += len(m.group(0)) - 1  # KEEP THE LAST NEWLINE.
                else:
                    break
            m = self._spaces.match(self.input, self.pos)
            log.trace(
                "Checking in/dedent, m=%s (length=%s)."
                % (m, "n/a" if m is None else len(m.group(1)))
            )
            if m:
                indent = len(m.group(1))
                log.trace("indent=%u." % indent)
                if indent > self._indent[-1]:
                    self._indent.append(indent)
                    token = yapps.runtime.Token(
                        "INDENT",
                        indent,
                        self.get_pos(),
                    )
                    log.trace("token=%s." % (token,))
                    return token
                if indent < self._indent[-1]:
                    self._indent.pop()
                    token = yapps.runtime.Token(
                        "DEDENT",
                        indent,
                        self.get_pos(),
                    )
                    log.trace("token=%s." % (token,))
                    return token
        if "INDENTED_CODE" in restrict:
            m = self._indented_code.match(self.input, self.pos)
            log.trace("Checking indented code, m=%s." % (m,))
            if m:
                indent = len(m.group(1))
                code = m.group(2)
                log.trace(
                    "indent=%u (was %u), code=%s."
                    % (
                        indent,
                        self._indent[-1],
                        code,
                    )
                )
                if indent >= self._indent[-1]:
                    code_spaces = " " * (indent - self._indent[-1])
                    line = code_spaces + code
                    token = yapps.runtime.Token(
                        "INDENTED_CODE",
                        line,
                        self.get_pos(),
                    )
                    log.trace("token=%s." % (token,))
                    self.pos += len(m.group(0))
                    return token
        token = super(Scanner, self).token(restrict, context)
        log.trace("token=%s." % (token,))
        return token


class Parser(state_machine):  # noqa: F821
    def __init__(self, *args, **kwargs):
        super(Parser, self).__init__(*args, **kwargs)

    def parse(self):
        spec = Specification()
        return self.machine_spec(spec)


class SyntaxError(Exception):
    pass


class SmaxException(Exception):
    """
    Used for semantic errors in the state machine spec.
    """

    pass


class Transition(object):
    def __init__(self, state, event, superclasses, condition, target, code):
        self.state = state
        self.event = event
        self.condition = condition
        self.target = target
        self.code = code
        self.superclasses = superclasses

    def find(self, state):
        return "target=%s/%s." % (state.name, self.target)


class Timeout(object):
    def __init__(self, state, time_spec, condition, target, code):
        self.state = state
        self.time_spec = time_spec
        self.condition = condition
        self.target = target
        self.code = code


class TimeSpec(object):
    def __init__(self, expr, scale):
        self.timeout = expr
        self.scale = scale


class Event(object):
    def __init__(self, event, event_args, superclasses):
        self.name = event
        self.args = event_args
        self.superclasses = []
        self._superclasses = {}
        self.merge_superclasses(superclasses)

    def merge_superclasses(self, superclasses):
        for superclass in superclasses:
            name, args = superclass
            existing_args = self._superclasses.get(name, None)
            if existing_args:
                if len(args) != len(existing_args):
                    raise SmaxException(
                        "Incompatible parameter list with %s superclass %s."
                        % self.name,
                        name,
                    )
                continue
            self.superclasses.append(superclass)
            self._superclasses[name] = args


class State(object):
    def __init__(self, machine, parent, name, start):
        self._machine = machine
        self.parent = parent
        self.name = name
        self.start = start
        self.enter = []
        self.exit = []
        self.timeouts = []
        self.transitions = []
        self.inner_states = []
        self._all_inner_states = []
        self._state = {}
        self._events = {}
        self.events = []
        self._default_transition = None

    def check(self, machine):
        # make sure all our inner states have exactly one start state.
        for sl in self.inner_states:
            start_found = False
            for s in sl:
                if s.start and start_found:
                    raise SmaxException(
                        "Multiple start states found in %s." % self.name,
                    )
                if s.start:
                    start_found = True
            if not start_found:
                raise SmaxException("No start states found in %s." % self.name)
        u = [self.name]
        d = [self.name]
        s = self
        while s.parent:
            p = s.parent
            u.insert(0, p.name)
            d.insert(0, p.name)
            for n, l in enumerate(p.inner_states):
                if s in l:
                    u.insert(1, "%d" % n)
                    break
            s = p
        self.full_name = "_".join(u)
        self.dot_name = ".".join(d)
        self.name_list = d
        self.array_name = self.full_name + "_name"
        log.trace(
            "%s: full_name=%s, dot_name=%s."
            % (
                self,
                self.full_name,
                self.dot_name,
            )
        )
        target_name = "n/a"
        n = 0  # in case transitions is empty
        for n, t in enumerate(self.transitions):
            log.trace("t=%s, target=%s." % (t, t.target))
            t.n = n
            if t.target is None:
                continue
            m = self
            i = 0
            while t.target[i] == "^":
                if not m.parent:
                    raise SmaxException("Cannot go up from %s." % (m.name,))
                m = m.parent
                log.trace("t.target[%d] == '^'; going up to %s." % (i, m.name))
                i += 1
            # assert len(t.target)==(i + 1)
            try:
                # Is the target state one of our children?
                # If so then don't unconfigure ourselves.
                q = m._state[t.target[i]]
                t.unconfigure = False
            except KeyError:
                # Is the target state one of our parent's children?
                if not m.parent:
                    raise SmaxException("Can't find target state %s." % (t.target[i],))
                q = m.parent._state[t.target[i]]
                t.unconfigure = True
            log.trace("t.target[%d] found %s." % (i, q.name))
            i += 1
            while i < len(t.target):
                q = q._state[t.target[i]]
                log.trace("t.target[%d] found %s." % (i, q.name))
                i += 1
            t.target_state = q
            target_name = q.name
        for n, t in enumerate(self.timeouts):
            log.trace("t=%s, target=%s." % (t, target_name))
            t.n = n
            if t.target is None:
                continue
            m = self.parent
            i = 0
            while t.target[i] == "^":
                m = m.parent
                i += 1
            assert len(t.target) == (i + 1)
            t.target_state = m._state[t.target[i]]
            t.unconfigure = True

    def _full_name(self):
        if self.parent:
            r = self.parent._full_name()
            if self.parent.inner_states:
                for n, l in self.parent.inner_states:
                    if self in l:
                        r.append("%d", n)
                        break
        r.append(self.name)
        return r

    def set_enter(self, enter):
        self.enter = enter

    def set_exit(self, exit):
        self.exit = exit

    def add_transition(
        self,
        event_name,
        event_args,
        superclasses,
        condition,
        state_target,
        code_clause,
    ):
        ev = self._machine.event(event_name, event_args, superclasses)
        t = Transition(
            self,
            ev,
            superclasses,
            condition,
            state_target,
            code_clause,
        )
        self.transitions.append(t)
        self.add_event(ev)

    def default_transition(self, condition, state_target, code_clause):
        t = Transition(self, None, [], condition, state_target, code_clause)
        if condition is None:
            if self._default_transition:
                raise SyntaxError(
                    "State %s has multiple default transitions." % (self.name,)
                )
            self._default_transition = t
        self.transitions.append(t)

    def add_event(self, ev):
        self._events[ev] = ev
        self.events.append(ev)
        if self.parent:
            self.parent.add_event(ev)

    def new_context(self):
        return self

    def timeout_s(self, expr):
        return TimeSpec(expr, "s")

    def timeout_ms(self, expr):
        return TimeSpec(expr, "ms")

    def add_timeout(self, time_spec, condition, state_target, code_clause):
        t = Timeout(self, time_spec, condition, state_target, code_clause)
        self.timeouts.append(t)

    def state(self, name, start):
        state_name = name
        log.trace("state name=%s." % (state_name,))
        try:
            # is "state_name" in self._state?
            self._state[state_name]
            raise SyntaxError("State %s is duplicate." % (state_name,))
        except KeyError:
            pass
        s = State(self._machine, self, name, start)
        log.trace("new state=%s, parent=%s." % (s.name, s.parent.name))
        self._state[state_name] = s
        return s

    def state_machine(self, states):
        n = len(self.inner_states)
        for i, s in enumerate(states):
            s.n = n
            s.or_with = states
            s.or_n = i
        self.inner_states.append(states)
        self._all_inner_states.extend(states)

    def all_states_depth_first(self):
        if not self._all_inner_states:
            return [self]
        r = []
        for m in self.inner_states:
            for i in m:
                r.extend(i.all_states_depth_first())
        r.append(self)
        return r

    # breadth first
    def all_states(self):
        r = [self]
        for m in self.inner_states:
            for i in m:
                r.extend(i.all_states())
        return r


#   def start_states(self):
#       if not self.start:
#           return [ ]
#       r = [ self ]
#       for i in self._all_inner_states:
#           r.extend(i.start_states())
#       return r


class Machine(State):
    def __init__(self, name, superclass):
        super(Machine, self).__init__(self, None, name, True)
        self.superclass = superclass
        self._event = {}

    def context(self):
        return self

    def event(self, event, event_args, superclasses):
        event_name = event
        log.trace(
            "event_name=%s, event_args=%s, superclasses=%s."
            % (event_name, event_args, superclasses)
        )
        try:
            ev = self._event[event_name]
            if len(event_args) and (len(event_args) != len(ev.args)):
                raise SyntaxError(
                    "Event %s with different argument list.",
                    event_name,
                )
            ev.merge_superclasses(superclasses)
        except KeyError:
            # make a new event.
            ev = Event(event, event_args, superclasses)
            self._event[event_name] = ev
        return ev

    def check(self, machine=None):
        self.event_list = list(self._event.values())
        self.event_list.sort(key=lambda ev: ev.name)
        super(Machine, self).check(self)
        for s in self.all_states():
            if s != self:
                s.check(self)
        # go through all the transitions and find
        # their targets.


#       for m in self.inner_states:
#           for s in m:
#               for t in s.transitions:
#                   log.debug("t=%s, target=%s." % (t, t.target))
#                   if t.target == None:
#                       continue
#                   # log.debug("self._state=%s." % (self._state.keys(),))
#                   assert len(t.target)==1
#                   t.target_state = self._state[t.target[0]]
#               for t in s.timeouts:
#                   log.debug("t=%s, target=%s." % (t, t.target))
#                   if t.target == None:
#                       continue
#                   # log.debug("self._state=%s." % (self._state.keys(),))
#                   assert len(t.target)==1
#                   t.target_state = self._state[t.target[0]]


class Macro(Machine):
    def __init__(self, name):
        super(Macro, self).__init__(name, "object")


#       self._name = name
#       self.enter = [ ]
#       self.exit = [ ]
#   def context(self):
#       return self
#   def set_enter(self, enter):
#       self.enter = enter
#   def set_exit(self, exit):
#       self.exit = exit


class Specification:
    def __init__(self):
        log.trace("new Specification.")
        self._output = []

    def constant(self, name, value):
        log.trace("new constant, name=%s, value=%s." % (name, value))
        self._output.append({"constant": {"name": name, "value": value}})

    def import_(self, sequence):
        log.trace("new import, sequence=%s." % (sequence,))
        self._output.append({"import": sequence})

    def machine(self, name, superclass):
        m = Machine(name, superclass)
        self._output.append({"machine": m})
        return m

    def spec(self):
        return self._output

    def macro(self, name):
        m = Macro(name)
        self._output.append({"macro": m})
        return m



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import os
import select
import smax

import smax.log as log


class SelectReactor(smax.Reactor):
    """
    Reactor that works with unix's select so that the entire application
    works in a single thread.  Use add_fd and remove_fd to trigger
    a callback when select() returns data ready on the file descriptor.
    """

    update = b"U"

    def __init__(self):
        self._control_read, self._control_write = os.pipe()
        self._r = {self._control_read: self.__control_ready}
        self._w = {}
        self._x = {}
        super(SelectReactor, self).__init__()

    def run(self):
        while True:
            timeout = self.sync()
            if self.done():
                return
            # timeout may be None
            log.trace("timeout=%s." % timeout)
            r, w, x = select.select(
                self._r.keys(), self._w.keys(), self._x.keys(), timeout
            )
            for ir in r:
                self._r[ir]()
            for iw in w:
                self._w[iw]()
            for ix in x:
                self._x[ix]()

    def _signal(self):
        os.write(self._control_write, self.update)

    def add_fd(self, fd, read_callback):
        self._r[fd] = read_callback
        self._signal()

    def remove_fd(self, fd):
        del self._r[fd]
        self._signal()

    def __control_ready(self):
        # Get 'update' out of the ingress port.
        os.read(self._control_read, 4096)



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

# Reactor that works with PyQt5 so that the entire application
# works in a single thread.

try:
    from PyQt5 import QtCore, QtGui, QtWidgets
except ModuleNotFoundError:
    from PySide2 import QtCore, QtGui, QtWidgets

import smax
import smax.log as log


class PyQtReactor(smax.Reactor):
    class Adapter(QtCore.QObject):
        _queue_update = QtCore.pyqtSignal(object)

        def __init__(self, reactor):
            super(PyQtReactor.Adapter, self).__init__()
            self._reactor = reactor
            self._o = []  # objects we want to keep references to
            self._queue_update.connect(self.execute)
            self._timer = QtCore.QTimer()
            self._timer.timeout.connect(self.timeout)
            self._timer.setSingleShot(True)

        def execute(self, cb):
            self._timer.stop()
            cb()

        def signal(self, cb):
            self._queue_update.emit(cb)

        def schedule(self, timeout, cb):
            self._timer.stop()
            self.timeout_cb = cb
            self._timer.start(timeout * 1000.0)

        def timeout(self):
            self.timeout_cb()

        def add_fd(self, fd, read_callback):
            notifier = QtCore.QSocketNotifier(fd, QtCore.QSocketNotifier.Read)
            notifier.activated.connect(read_callback)
            self._o.append(notifier)

        def remove_fd(self, fd):
            for n, o in enumerate(
                self._o[:]
            ):  # iterate over a copy so we can delete the element
                if o.socket() == fd:
                    self._o.pop(n)
                    return

    def __init__(self):
        super(PyQtReactor, self).__init__()
        self._adapter = PyQtReactor.Adapter(self)

    def _signal(self):
        self._adapter.signal(self._run)

    def _run(self):
        timeout = self.sync()
        if self.done():
            return
        # timeout may be None
        log.trace("timeout=%s." % timeout)
        if timeout is not None:
            self._adapter.schedule(timeout, self._run)

    def add_fd(self, fd, read_callback):
        self._adapter.add_fd(fd, read_callback)

    def remove_fd(self, fd):
        self._adapter.remove_fd(fd)


def machine(spec, machine_name):
    for m in spec["spec"]:
        if "machine" not in m:
            continue
        machine = m["machine"]
        if machine.name != machine_name:
            continue
        return machine
    return None


NEST = "NEST"
UNNEST = "UNNEST"


def states(state):
    yield state
    if not state.inner_states:
        return
    yield NEST
    for sl in state.inner_states:
        for s in sl:
            for ss in states(s):
                yield ss
    yield UNNEST


def state_tree(machine, tree_widget):
    r = {}
    parent = [tree_widget]
    last = None
    items = []
    for s in states(machine):
        if s == NEST:
            parent.insert(0, last)
            continue
        if s == UNNEST:
            parent.pop(0)
            continue
        item = QtWidgets.QTreeWidgetItem(parent[0], [s.name])
        r[s.dot_name] = item
        last = item
        items.append(item)
    tree_widget.addTopLevelItems(items)
    for item in items:
        item.setExpanded(True)
        item.setCheckState(0, QtCore.Qt.Unchecked)
        item.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
    return r


class EventWidget(QtWidgets.QWidget):
    def __init__(self, instance, event, parent=None):
        super(EventWidget, self).__init__(parent)
        self._event = event
        self._layout = QtGui.QHBoxLayout(self)
        self._control = QtGui.QPushButton(event.name, self)
        self._layout.addWidget(self._control)
        parameters = {}
        for arg in event.args:
            text_edit = QtGui.QLineEdit(self)
            self._layout.addWidget(QtGui.QLabel(arg))
            self._layout.addWidget(text_edit)
            parameters[arg] = text_edit
        method = getattr(instance, event.name)

        def go(state, method=method, event=event, parameters=parameters):
            args = [str(parameters[arg].text()) for arg in event.args]
            method(*args)

        self._control.clicked.connect(go)
        self.setLayout(self._layout)


def event_controls(machine, instance, scroll_area):
    events = {e.name: e for e in machine.events}
    keys = sorted(events.keys())
    for k in keys:
        ew = EventWidget(instance, events[k], scroll_area)
        list_item = QtGui.QListWidgetItem(scroll_area)
        list_item.setSizeHint(ew.sizeHint())
        scroll_area.addItem(list_item)
        scroll_area.setItemWidget(list_item, ew)



#

import os
from setuptools import setup
import sys

# Yapps doesn't seem to install
# properly without this.
sys.dont_write_bytecode = True

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.md"), "rt") as f:
    long_description = f.read()
long_description_content_type = "text/markdown"

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, "smax", "__version__.py")) as f:
    exec(f.read(), about)

setup(
    name="smax_dsl",
    description="smax: a DSL for Harel state machines in Python.",
    author="Patrick O'Grady",
    author_email="patrick.ogrady.gm@gmail.com",
    url="https://github.com/baymotion/smax",
    version=about["__version__"],
    packages=["smax"],
    install_requires=[
        "yapps",
        "jinja2",
        "pyyaml",
    ],
    entry_points={
        "console_scripts": [
            "smax=smax.__main__:main",
        ],
    },
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    test_suite="tests",
    license="GPLv3",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import os
import time
import traceback

start = time.time()


class NoTranscript:
    def write(self, msg):
        pass

    def flush(self):
        pass


transcript = NoTranscript()
enable_trace = False
enable_debug = True


def caller(n=0):
    tb = traceback.extract_stack(limit=3 + n)
    full_filename, line, method, statement = tb[-(3 + n)]
    filename = os.path.basename(full_filename)
    return "%s:%u" % (filename, line)


def write(enable, s):
    transcript.write("%s\n" % s)
    transcript.flush()
    if enable:
        print(s)


def trace(msg):
    write(
        enable_trace,
        "TRACE %u %.2lf %s -- %s"
        % (
            os.getpid(),
            time.time() - start,
            caller(),
            msg,
        ),
    )


def _trace(msg):
    """
    _trace is the same as trace except that it reports
    the caller as the one above who called _trace.
    """
    write(
        enable_trace,
        "TRACE %u %.2lf %s -- %s"
        % (
            os.getpid(),
            time.time() - start,
            caller(1),
            msg,
        ),
    )


def debug(msg):
    write(
        enable_debug,
        "DEBUG %u %.2lf %s -- %s"
        % (
            os.getpid(),
            time.time() - start,
            caller(),
            msg,
        ),
    )


def _debug(msg):
    """
    _debug is the same as trace except that it reports
    the caller as the one above who called _debug.
    """
    write(
        enable_debug,
        "DEBUG %u %.2lf %s -- %s"
        % (
            os.getpid(),
            time.time() - start,
            caller(1),
            msg,
        ),
    )


def error(msg):
    write(
        True,
        "ERROR %u %.2lf %s -- %s"
        % (
            os.getpid(),
            time.time() - start,
            caller(),
            msg,
        ),
    )


def _error(msg):
    """
    _error is the same as trace except that it reports the caller as the one
    above who called _error.
    """
    write(
        True,
        "ERROR %u %.2lf %s -- %s" % (os.getpid(), time.time() - start, caller(1), msg),
    )



# This file is part of the smax project (http://github.com/pjogrady/smax)
# and is copyrighted under GPL v3 or later.

import queue
import time

import smax.log as log


# Reactor framework
class Reactor(object):
    def __init__(self):
        super(Reactor, self).__init__()
        self._q = queue.Queue()
        self._alarms = []
        self._done = False

    # run the reactor until all queued and expired
    # events are done; returns a timeout in seconds
    # until the next event, or None if no alarms are active.
    def sync(self):
        while not self.done():
            if not self._q.empty():
                cb, args = self._q.get()
                log.trace("execute cb=%s." % cb)
                cb(*args)
                continue
            timeout = None
            now = time.time()
            if self._alarms:
                trigger, cb, args = self._alarms[0]
                if trigger <= now:
                    self._alarms.pop(0)
                    log.trace("alarm cb=%s." % cb)
                    cb(*args)
                    continue
                # we've reached our next closest timeout.
                timeout = trigger - now
            return timeout

    def call(self, cb, *args):
        log.trace("queue cb=%s." % cb)
        self._q.put((cb, args))
        self._signal()

    def after_s(self, seconds, callback, *args):
        trigger = time.time() + seconds
        r = (trigger, callback, args)
        log.trace("after_s cb=%s." % callback)
        self._alarms.append(r)
        self._alarms.sort()
        self._signal()
        return r

    def after_ms(self, ms, callback, *args):
        return self.after_s(ms / 1000.0, callback, *args)

    def cancel_after(self, r):
        if r not in self._alarms:
            return
        self._alarms.remove(r)
        self._signal()

    def done(self):
        return self._done

    def stop(self):
        log.trace("stop")
        self._done = True
        self._signal()

    def _signal(self):
        assert False

    def _run_event(self, machine, ev):
        ev(machine)
